k// parse.guile.c
// (C) 1998 lars brinkhoff.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "lasr.h"

Object *objects[100];

struct
{
  char *name;
  Material *pointer;
} materials[10];

Sky sky;

Real sky_day (Sky *sky, Vector *d)
{
  return 0.5 * vector_scalar (sky->up, *d) + 0.5;
}

Real sky_black (Sky *sky, Vector *d)
{
  return 0;
}

void check_f (Object *o, Vector *p, Vector *n, Material *m)
{
  static const Material black = (Material){ 0, 0.1, 0, 0.1, 0 };
  static const Material white = (Material){ 0, 0.1, 0.9, 0.1, 0 };

  if (((int)(0.05 * p->x + 1e6) & 1) ^ ((int)(0.05 * p->z + 1e6) & 1))
    *m = black;
  else
    *m = white;
}

Material *create_material (Real e, Real s, Real d, Real b)
{
  Material *m;
  
  m = safe_malloc (sizeof (Material));
  
  m->emit = e;
  m->spec = s;
  m->diff = d;
  m->grain = b;
  m->f = 0;

  return m;
}

void
skip_whitespace (void)
{
  int c;

  while ((c = getchar ()) != EOF &&
	 index (" \t\n", c) != 0)
    ;
  
  ungetc (c, stdin);
}

int
parse_int (void)
{
  int c;
  int i = 0;
  int negative = 0;
  
  skip_whitespace ();

  c = getchar ();
  if (c == '-')
    negative = 1;
  else
    ungetc (c, stdin);

  while ((c = getchar ()) != EOF &&
	 index ("0123456789", c) != 0)
    i = 10 * i + c - '0';

  ungetc (c, stdin);
  
  if (negative)
    return -i;
  else
    return i;
}  

Real
parse_real (void)
{
  Real r;
  int c;

  r = parse_int ();

  c = getchar ();
  if (c == '.')
    {
      Real x = 0.1;
      
      while ((c = getchar ()) != EOF &&
	     index ("0123456789", c) != 0)
	{
	  r += x * (c - '0');
	  x *= 0.1;
	}
    }

  ungetc (c, stdin);

  return r;
}

void
expect_char (int d)
{
  int c;

  c = getchar ();
  if (c != d)
    {
      fprintf (stderr, "'%c' expected instead of '%c'\n", d, c);
      exit (1);
    }
}

Vector
parse_vector (void)
{
  Vector v;
  
  skip_whitespace ();
  expect_char ('(');
  v.x = parse_real ();
  skip_whitespace ();
  expect_char (',');
  v.y = parse_real ();
  skip_whitespace ();
  expect_char (',');
  v.z = parse_real ();
  skip_whitespace ();
  expect_char (')');

  return v;
}

char *
parse_identifier (void)
{
  char id[100], *id_;
  int i = 0;
  int c;

  skip_whitespace ();
  
  if ((c = getchar ()) != EOF &&
      index ("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ", c) != 0)
    {
      id[i++] = c;
      
      while ((c = getchar ()) != EOF &&
	     index ("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
		    "0123456789", c) != 0)
	if (i < 98)
	  id[i++] = c;
    }
  id[i] = 0;
  
  ungetc (c, stdin);
  
  id_ = strdup (id);
  if (id_ == 0)
    {
      fprintf (stderr, "out of memory\n");
      exit (1);
    }

  return id_;
}

Material *
parse_material (void)
{
  char *n;
  int i;

  n = parse_identifier ();

  for (i = 0; materials[i].pointer != 0; i++)
    if (strcmp (n, materials[i].name) == 0)
      {
	free (n);
	return materials[i].pointer;
      }
  
  free (n);
  return 0;
}

#include <fcntl.h>

void
parse_input (void)
{
  int i = 0;
  int j = 0;
  int c;

  /*
  Material light100_mat = (Material){ 100, 0, 0, 0, 0 };
  Material light30_mat = (Material){ 30, 0, 0, 0, 0 };
  Material light10_mat = (Material){ 10, 0, 0, 0, 0 };
  Material light3_mat = (Material){ 3, 0, 0, 0, 0 };
  Material light1_mat = (Material){ 1, 0, 0, 0, 0 };
  Material mirror_mat = (Material){ 0, 1, 0, 0, 0 };
  Material blurred_mirror_mat = (Material){ 0, 1, 0, 0.2, 0 };
  Material matte_white_mat = (Material){ 0, 0, 1, 0, 0 };
  Material checkerboard_mat = (Material){ 0, 0, 0, 0, check_f };
  */

#if 1
    sky.f = 0;
#else
    sky.f = sky_day;
    sky.up = vec (0, 1, 0);
#endif

  objects[0] = 0;
  materials[0].pointer = 0;

  while (!feof (stdin))
    {
      skip_whitespace ();
      c = getchar ();

      switch (c)
	{
	case 's':
	  {
	    Material *m;
	    Vector p;
	    Real r;
	    
	    p = parse_vector ();
	    r = parse_real ();
	    m = parse_material ();

	    objects[i++] = create_sphere (p, r, m, &E);
	  }
	break;
	case 'r':
	  {
	    Material *m;
	    Vector p, n;
	    Real r1, r2;
	    
	    p = parse_vector ();
	    n = parse_vector ();
	    r1 = parse_real ();
	    r2 = parse_real ();
	    m = parse_material ();

	    objects[i++] = create_ring (p, n, r1, r2, m, &E);
	  }
	break;
	case 'M':
	  {
	    Real e, s, d, b;
	    Material *m;
	    char *n;
	    
	    n = parse_identifier ();
	    e = parse_real ();
	    s = parse_real ();
	    d = parse_real ();
	    b = parse_real ();
	    
	    m = create_material (e, s, d, b);

	    materials[j].name = n;
	    materials[j++].pointer = m;

	    materials[j].pointer = 0;
	  }
	break;
	case 'R':
	  {
	    int width, height, rpp;

	    width = parse_int ();
	    height = parse_int ();
	    rpp = parse_int ();
	    
	    render (width, height, rpp);
	    
	    return;
	  }
	break;
	default:
	  fprintf (stderr, "unrecognized entity: %c\n", c);
	}
      objects[i] = 0;
    }

#if 0
  if (0)
    {
      sky.f = sky_black;
      sky.up = vec (0, 1, 0);

      objects[0] = create_sphere (vec (0.0, 30.0, 0.0), 40.0,
				  &matte_white_mat, &E);
      
      options.N += 4;
      
      for (i = 1; i < 11; i++)
	{
	  Real x, z;
	  
	  x = 80 * cos (2 * pi * i / 10.0 +
			2 * pi / 10.0 / 16.0 * options.N);
	  z = 80 * sin (2 * pi * i / 10.0 +
			2 * pi / 10.0 / 16.0 * options.N);
	  objects[i] = create_sphere (vec (x, -20.0, z), 15.0,
				      &matte_white_mat, &E);
	}
      
      objects[i++] = create_sphere (vec (60.0, 60.0, -30.0), 15.0,
				    &light30_mat, &E);
      objects[i++] = create_sphere (vec (45, 75, -40), 5,
				    &matte_white_mat, &E);
      
      objects[i++] = create_ring (vec (60, 60, -30),
				  vec (2, 4, -1),
				  25, 20,
				  &matte_white_mat, &E);

      objects[i++] = create_ring (vec (0, -35, 0.0),
				  vec (0, 1, 0),
				  100, 60,
				  &matte_white_mat, &E);
    }
  else if (0)
    {
      Real r = 15.0;
      int a, b, c;
      Matrix m;
      
      m = matrix_multiply (matrix_rot_x (-0.2 * pi), 
			   matrix_rot_y (0.2 * pi));
      
      for (a = 4; a > 0; a--)
	for (b = a; b > 0; b--)
	  for (c = a; c > 0; c--)
	    objects[i++] = create_sphere (xv (m, r * (2 * b - a - 1),
					      r * (2 - a),
					      r * (2 * c - a - 1)), r,
					  &mirror_mat,
					  &E);
      
      
      m = matrix_rot_x (-0.2 * pi);
      
      objects[i++] = create_ring (xv (m, 0, (r / sqrt (2)) * -2 - r, 0),
				  xv (m, 0, 1, 0), 1000, 0,
				  &checkerboard_mat,
				  &E);
      
      objects[i++] = create_sphere (xv (m, 200, 200, -200), 150, &light10_mat,
				    &E);
      
      objects[i++] = create_sphere (xv (m, 0, 300, 0), 200, &light10_mat, &E);
      
      objects[i++] = create_sphere (xv (m, -300 + options.N * 30, 0, -200), 80,
				    &mirror_mat, &E);
    }
  else if (0)
    {
      // A motion-blurred rind spinning half a revolution.

      Matrix m = matrix_rot_y (options.N * 0.1 * pi);
      Matrix Dm = matrix_rot_y (0.1 * pi);
      //Matrix m = matrix_translate (vec (30, 0, 0));
      //Matrix m = matrix_multiply (matrix_rot_y (0.5 * pi),
      //			matrix_translate (vec (60, 0, 0)));
      objects[i++] = create_ring (xv (m, 0, 0, 0), xv (m, 0, 0, -1), 30, 50,
				  &light1_mat, &Dm);
    }
  else if (0)
    {
      // Three stacked white balls on top of a blurred mirror.

      Matrix m;

      m = matrix_multiply (matrix_translate (vec (0, 0, 0)),
			   matrix_rot_x (options.N * -0.01 * pi));

      objects[i++] = create_sphere (xv (m, 0, 0, 0), 20,
				    &mirror_mat, &E);
      objects[i++] = create_sphere (xv (m, 0, 40, 0), 20,
				    &mirror_mat, &E);
      objects[i++] = create_sphere (xv (m, 0, 80, 0), 20,
				    &mirror_mat, &E);
      objects[i++] = create_ring (xv (m, 0, -20, 0),
				  xv (m, 0, 1, 0), 0, 60,
				  &blurred_mirror_mat, &E);
      /*
      objects[i++] = create_sphere (xv (m, 100, 200, 0), 100,
				    &light1_mat, &E);
      objects[i++] = create_sphere (xv (m, -100, 200, 0), 100,
				    &light10_mat, &E);
				    */
      sky.f = sky_day;
      sky.up = xv (m, 0, 1, 0);
    }
  else if (0)
    {
      objects[i++] = create_ring (vec (-50, 0, 0), vec (0, 1, -1), 0, 40,
				  &mirror_mat, &E);
      objects[i++] = create_tri (vec (10, 0, 0),
				 vec (90, 90, -0),
				 vec (90, -90, 0),
				 vec (1, 0, -1),
				 vec (0, -1, -1),
				 vec (0, 1, -1),
				 &mirror_mat, &E);
    }
  else if (0)
    {
      sky.f = sky_black;

      // floor
      objects[i++] = create_ring (vec (0, -70, 0),
				  vec (0, 1, 0),
				  0, 500,
				  &matte_white_mat, &E);
      // roof
      objects[i++] = create_ring (vec (0, 150, 0),
				  vec (0, -1, 0),
				  0, 500,
				  &matte_white_mat, &E);
      // left wall
      objects[i++] = create_ring (vec (-100, 0, 0),
				  vec (1, 0, 0),
				  0, 500,
				  &matte_white_mat, &E);
      // right wall
      objects[i++] = create_ring (vec (100, 0, 0),
				  vec (-1, 0, 0),
				  0, 500,
				  &matte_white_mat, &E);
      // back wall
      objects[i++] = create_ring (vec (0, 0, 100),
				  vec (0, 0, -1),
				  0, 500,
				  &matte_white_mat, &E);
      // ball floating above floor
      objects[i++] = create_sphere (vec (20, 0, 0), 40,
				    &matte_white_mat, &E);
      // light source
      objects[i++] = create_ring (vec (100, 100, 0),
				  vec (-1, -1, 0),
				  0, 500,
				  &light10_mat, &E);
    }
#endif
  
  //objects[i] = 0;
  //render (options.width, options.height, options.rays_per_pixel);
  // render (80, 80, 8);
}
