/*

ring.c

(C) 1997, 1998 lars brinkhoff

*/

#include <stdio.h>
#include <stdlib.h>

#include "safe.h"
#include "objects.h"

struct Ring
{
  Object o;
  Vector p, n;
  Real rmax2, rmin2;
};

Real
ring_intersection (void *ring_, Ray *ray)
{
  struct Ring *ring = ring_;
  Real alpha;
  Real A, q2;
  Vector p, n;
  Vector v;
  Vector q;
  Matrix M;

  matrix_ptr_interpolate_with_E (&M, ray->t, ring->o.motion);
  vector_ptr_transform (&p, &M, &ring->p);
  vector_ptr_transform_no_translation (&n, &M, &ring->n);

  A = vector_scalar (ray->d, n);
  if (A == 0.0)
    return 0.0;
  
  v.x = p.x - ray->p.x;
  v.y = p.y - ray->p.y;
  v.z = p.z - ray->p.z;
  alpha = vector_scalar (v, n) / A;
  
  q.x = alpha * ray->d.x - v.x;
  q.y = alpha * ray->d.y - v.y;
  q.z = alpha * ray->d.z - v.z;

  q2 = vector_scalar (q, q);
  if (ring->rmin2 <= q2 && q2 <= ring->rmax2)
    return alpha;

  return 0.0;
}

void
ring_normal (Vector *n, void *ring_, Vector *p, Real t)
{
  struct Ring *ring = ring_;
  Matrix M;

  matrix_ptr_interpolate_with_E (&M, t, ring->o.motion);
  vector_ptr_transform (n, &M, &ring->n);
}

Object *
create_ring (Vector p, Vector n, Real r1, Real r2,
	     const Material *m, const Matrix *motion)
{
  struct Ring *ring = safe_malloc (sizeof (struct Ring));
  
  if (r1 < r2)
    {
      Real tmp = r2;
      r2 = r1;
      r1 = tmp;
    }
  ring->o = create_object (ring_intersection,
			   ring_normal,
			   m,
			   motion);
  ring->p = p;
  ring->n = n; vector_normalize (ring->n);
  ring->rmax2 = r1 * r1;
  ring->rmin2 = r2 * r2;
  
  return (Object *)ring;
}
