/*

sphere.c

(C) 1997, 1998 lars brinkhoff

*/

#include <stdio.h>
#include <stdlib.h>

#include "safe.h"
#include "objects.h"

struct Sphere
{
  Object o;
  Vector p;
  Real r;
};

Real
sphere_intersection (void *sphere_, Ray *ray)
{
  struct Sphere *sphere = sphere_;
  Real D, E, F, G;
  Real alpha;
  Matrix M;
  Vector v;
  Vector p;

  matrix_ptr_interpolate_with_E (&M, ray->t, sphere->o.motion);
  vector_ptr_transform (&p, &M, &sphere->p);

  v.x = ray->p.x - p.x;
  v.y = ray->p.y - p.y;
  v.z = ray->p.z - p.z;
  D = vector_scalar (v, ray->d);
  E = vector_scalar (v, v) - sphere->r * sphere->r;
  F = D * D - E;
  if (F < 0.0)
    return 0.0;
  G = sqrt (F);
  alpha = -D - G;
  if (alpha > 0.0)
    return alpha;
  alpha = -D + G;
  if (alpha > 0.0)
    return alpha;
  return 0.0;
}

void
sphere_normal (Vector *n, void *sphere_, Vector *p, Real t)
{
  struct Sphere *sphere = sphere_;
  Matrix M;
  Vector q;

  matrix_ptr_interpolate_with_E (&M, t, sphere->o.motion);
  vector_ptr_transform (&q, &M, &sphere->p);

  n->x = p->x - q.x;
  n->y = p->y - q.y;
  n->z = p->z - q.z;
  vector_normalize (*n);
}

Object *
create_sphere (Vector p, Real r, const Material *m, const Matrix *motion)
{
  struct Sphere *sphere = safe_malloc (sizeof (struct Sphere));
  
  sphere->o = create_object (sphere_intersection,
			     sphere_normal,
			     m,
			     motion);
  sphere->p = p;
  sphere->r = r;
    
  return (Object *)sphere;
}
