/*

tri.c

(C) 1998 lars brinkhoff

*/

#include <stdio.h>
#include <stdlib.h>

#include "safe.h"
#include "objects.h"

struct Tri
{
  Object o;
  Vector p1, p2, p3;
  Vector n1, n2, n3;
  Vector n;
};

Real
tri_intersection (void *tri_, Ray *ray)
{
  struct Tri *tri = tri_;
  Real alpha;
  Real A;
  Vector p1, p2, p3;
  Vector n;
  Vector v;
  Matrix M;
  Vector D1, D2;
  Real beta, gamma;

  matrix_ptr_interpolate_with_E (&M, ray->t, tri->o.motion);
  vector_ptr_transform (&p1, &M, &tri->p1);
  vector_ptr_transform (&p2, &M, &tri->p2);
  vector_ptr_transform (&p3, &M, &tri->p3);
  vector_ptr_transform_no_translation (&n, &M, &tri->n);

  A = vector_scalar (ray->d, n);
  if (A == 0.0)
    return 0.0;
  
  v.x = p1.x - ray->p.x;
  v.y = p1.y - ray->p.y;
  v.z = p1.z - ray->p.z;
  alpha = vector_scalar (v, n) / A;
#if 1
  D1.x = p2.x - p1.x; D1.y = p2.y - p1.y; D1.z = p2.z - p1.z;
  D2.x = p3.x - p1.x; D2.y = p3.y - p1.y; D2.z = p3.z - p1.z;
  
  beta = (ray->p.x + alpha * ray->d.x - p1.x -
	  (ray->p.y + alpha * ray->d.y - p1.y) * D2.x / D2.y) /
         (D1.x - D1.y * D2.x / D2.y);
  gamma = (ray->p.y + alpha * ray->d.y - p1.y - beta * D1.y ) / D2.y;
#else
  {
    Vector q;
    Real aa, bb, ab, qa, qb;

    q.x = ray->p.x + alpha * ray->d.x;
    q.y = ray->p.y + alpha * ray->d.y;
    q.z = ray->p.z + alpha * ray->d.z;
    aa = vector_scalar (D1, D1);
    bb = vector_scalar (D2, D2);
    ab = vector_scalar (D1, D2);
    qa = vector_scalar (q, D1);
    qb = vector_scalar (q, D2);
    beta = (bb * qa - ab * qb) / (aa * bb - ab * ab);
    gamma = (qb - beta * ab) / bb;
  }
#endif
  if (beta < 0 || gamma < 0 || beta + gamma > 1)
    return 0.0;
  
  return alpha;
}

void
tri_normal (Vector *n, void *tri_, Vector *p, Real t)
{
  struct Tri *tri = tri_;
  Matrix M;
  Vector p1, p2, p3;
  Vector n1, n2, n3;
  Vector D1, D2, D3;
  Real r1, r2, r3;
  
  matrix_ptr_interpolate_with_E (&M, t, tri->o.motion);
  vector_ptr_transform (&p1, &M, &tri->p1);
  vector_ptr_transform (&p2, &M, &tri->p2);
  vector_ptr_transform (&p3, &M, &tri->p3);
  vector_ptr_transform_no_translation (&n1, &M, &tri->n1);
  vector_ptr_transform_no_translation (&n2, &M, &tri->n2);
  vector_ptr_transform_no_translation (&n3, &M, &tri->n3);

  D1.x = p1.x - p->x; D1.y = p1.y - p->y; D1.z = p1.z - p->z;
  D2.x = p2.x - p->x; D2.y = p2.y - p->y; D2.z = p2.z - p->z;
  D3.x = p3.x - p->x; D3.y = p3.y - p->y; D3.z = p3.z - p->z;

  r1 = vector_magnitude (D1);
  r2 = vector_magnitude (D2);
  r3 = vector_magnitude (D3);
  
  n->x = (r2 * r3 * n1.x + r1 * r2 * n2.x + r1 * r2 * n3.x) /
         (r2 * r3 + r1 * r3 + r1 * r2);
  n->y = (r2 * r3 * n1.y + r1 * r2 * n2.y + r1 * r2 * n3.y) /
         (r2 * r3 + r1 * r3 + r1 * r2);
  n->z = (r2 * r3 * n1.z + r1 * r2 * n2.z + r1 * r2 * n3.z) /
         (r2 * r3 + r1 * r3 + r1 * r2);
  
  vector_normalize (*n);
//fprintf (stderr, "%f %f %f %f %f %f\n", r1, r2, r3, n->x, n->y, n->z);
  matrix_ptr_interpolate_with_E (&M, t, tri->o.motion);
  vector_ptr_transform (n, &M, &tri->n);
}

Object *
create_tri (Vector p1, Vector p2, Vector p3,
	    Vector n1, Vector n2, Vector n3,
	    const Material *m, const Matrix *motion)
{
  struct Tri *tri = safe_malloc (sizeof (struct Tri));
  Vector a, b;
  
  tri->o = create_object (tri_intersection,
			  tri_normal,
			  m,
			  motion);
  tri->p1 = p1;
  tri->p2 = p2;
  tri->p3 = p3;
  tri->n1 = n1; vector_normalize (tri->n1);
  tri->n2 = n2; vector_normalize (tri->n2);
  tri->n3 = n3; vector_normalize (tri->n3);
  a.x = p2.x - p1.x; a.y = p2.y - p1.y; a.z = p2.z - p1.z;
  b.x = p3.x - p1.x; b.y = p3.y - p1.y; b.z = p3.z - p1.z;
  tri->n.x = a.y * b.z - a.z * b.y;
  tri->n.y = a.z * b.x - a.x * b.z;
  tri->n.z = a.x * b.y - a.y * b.x;
  vector_normalize (tri->n);
  
  return (Object *)tri;
}
