// lasrc/main.c
// (C) 1998 lars brinkhoff

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "lasrc.h"
#include "server.h"
#include "image_out_device.h"

void read_image (struct Options *, Image_out_device *, Image_out_device *);
struct Options *parse_options (int, char *[]);
Image_out_device *init_visual_output (char *, int, int, int, double);

extern char *optarg;
extern int optind, opterr, optopt;
extern getopt ();

int
main (int argc, char *argv[])
{
  Image_out_device *screen, *file = 0;
  struct Options *options;

  options = parse_options (argc, argv);
  screen = options->create_screen (0, options->width, options->height,
				   options->scale, options->gamma);
  if (options->outfile)
    file = create_rgb_image_out_device (options->outfile, options->width,
					options->height, options->scale,
					options->gamma);
  read_image (options, screen, file);
  if (file)
    file->close (file);
  if (screen)
    {
      screen->wait (screen);
      screen->close (screen);
    }

  return 0;
}

Server *
find_server_by_fd_in_a_really_slow_and_stupid_way (int fd, Server *s[])
{
  int i;

  for (i = 0; s[i] != 0; i++)
    if (server_fd(s[i]) == fd)
      return s[i];

  return 0;
}

void
read_image (struct Options *options,
	    Image_out_device *screen,
	    Image_out_device *file)
{
  int N = options->interleave;
  unsigned char buffer[1024];
  int answer;
  int x, y;
  int n;
  int i, j;

  Server *servers[50];

  char *hosts[] = { "localhost",
		    "scooter.mdstud.chalmers.se",
		    "dali1.mdstud.chalmers.se",
		    "dali2.mdstud.chalmers.se",
		    "dali3.mdstud.chalmers.se",
		    "dali4.mdstud.chalmers.se",
		    "dali5.mdstud.chalmers.se",
		    "dali6.mdstud.chalmers.se",
		    "dali7.mdstud.chalmers.se",
		    "dali8.mdstud.chalmers.se",
		    "dali9.mdstud.chalmers.se",
		    "dali10.mdstud.chalmers.se",
		    "dada1.mdstud.chalmers.se",
		    "dada2.mdstud.chalmers.se",
		    "dada3.mdstud.chalmers.se",
		    "dada4.mdstud.chalmers.se",
		    "dada5.mdstud.chalmers.se",
		    "dada6.mdstud.chalmers.se",
		    "dada7.mdstud.chalmers.se",
		    "dada8.mdstud.chalmers.se",
		    "dada9.mdstud.chalmers.se",
		    "dada10.mdstud.chalmers.se",
		    0 };

  fd_set fdset;

  for (i = j = 0; hosts[i] != 0; i++)
    {
      servers[j] = open_server (hosts[i], 10111, options->infile);
      if (servers[j] != 0)
	{
	  fprintf (stderr, "Connected to %s.\n", hosts[i]);
	  j++;
	}
    }
  servers[j] = 0;

  i = 0;
  for (y = 0; y < options->height; y += N)
    for (x = 0; x < options->width; x += N)
      {
	if (servers[i] == 0)
	  goto work;
	server_draw_area (servers[i], x, y, N, N,
			  options->width, options->height, options->rpp);
	i++;
      }

work:

  // Initialize fdset;
  FD_ZERO (&fdset);
  for (i = 0; servers[i] != 0; i++)
    FD_SET (server_fd (servers[i]), &fdset);

  while (y < options->height)
    {
      fd_set read_fd = fdset;
      select (50, &read_fd, 0, 0, 0);
      for (i = 0; servers[i] != 0; i++)
	if (FD_ISSET (server_fd (servers[i]), &read_fd))
	  {
	    n = read (server_fd (servers[i]), buffer, 1024);
fprintf (stderr, "%d bytes from server %d\n", n, i);
	    answer = server_receive_data (servers[i], screen, file, buffer, n);
	    if (answer == SERVER_DONE)
	      {
		server_draw_area (servers[i], x, y, N, N,
				  options->width, options->height,
				  options->rpp);
		x += N;
		if (x >= options->width)
		  {
		    x = 0;
		    y += N;
		  }
	      }
	  }
    }
}

struct Options *
parse_options (int argc, char *argv[])
{ 
  struct Options *options;
  int c;
  
  options = malloc (sizeof (struct Options));
  if (options == 0)
    {
      fprintf (stderr, "out of memory\n");
      exit (1);
    }

  /* Defaults. */
  options->create_screen = init_visual_output;
  options->width = -1;
  options->height = -1;
  options->scale = 1;
  options->interleave = 1;
  options->me = argv[0];
  options->infile = 0;
  options->outfile = 0;
  options->gamma = 1;
  options->rpp = 4;

  while ((c = getopt (argc, argv, "xvAs:w:h:i:g:r:f:")) != EOF)
    switch (c)
      {
#ifdef HAVE_X
      case 'x': options->create_screen = create_x_image_out_device; break;
#endif
#ifdef HAVE_AA
      case 'A': options->create_screen = create_aa_image_out_device; break;
#endif
      case 'v': options->create_screen = create_vt100_image_out_device; break;
      case 's': options->scale = atoi (optarg); break;
      case 'w': options->width = atoi (optarg); break;
      case 'h': options->height = atoi (optarg); break;
      case 'i': options->interleave = atoi (optarg); break;
      case 'g': options->gamma = atof (optarg); break;
      case 'r': options->rpp = atoi (optarg); break;
      case 'f': options->outfile = strdup (optarg);
      }

  if (optind < argc)
    options->infile = argv[optind++];
  else
    {
      fprintf (stderr, "%s: specify a file\n", options->me);
      exit (1);
    }

  if (optind < argc)
    {
      fprintf (stderr, "%s: only one file, pucko\n", options->me);
      exit (1);
    }

  if (options->width < 0)
    {
      fprintf (stderr,
	       "%s: raw input width must be specified.\n", options->me);
      exit (1);
    }

  if (options->create_screen != create_vt100_image_out_device &&
      options->height < 0)
    {
      fprintf (stderr,
	       "%s: input height must be specified when not using VT100.\n",
	       options->me);
      exit (1);
    }

  return options;
}

Image_out_device *
init_visual_output (char *name, int width, int height, int scale, double gamma)
{
  void *info;
  
#ifdef HAVE_X
  info = create_x_image_out_device (0, width, height, scale, gamma);
  if (info != 0)
    return info;
#endif
#ifdef HAVE_AA
  info = create_aa_image_out_device (0, width, height, scale, gamma);
  if (info != 0)
    return info;
#endif
  return create_vt100_image_out_device (0, width, height, scale, gamma);
}
