// lasrc/server.c
// (C) 1998 lars brinkhoff

// Handles communication with server.

#include <netdb.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "lasrc.h"
#include "server.h"
#include "image_out_device.h"

int
server_fd (Server *server)
{
  return server->socket;
}

/*                             Yes, I stole this code.  So?
int Socket(host, clientPort)
char *host;
int clientPort;
{
    int sock;
    unsigned long inaddr;
    struct sockaddr_in ad;
    struct hostent *hp;
    
    memset(&ad, 0, sizeof(ad));
    ad.sin_family = AF_INET;

    inaddr = inet_addr(host);
    if (inaddr != INADDR_NONE)
        memcpy(&ad.sin_addr, &inaddr, sizeof(inaddr));
    else
    {
        hp = gethostbyname(host);
        if (hp == NULL)
            return -1;
        memcpy(&ad.sin_addr, hp->h_addr, hp->h_length);
    }
    ad.sin_port = htons(clientPort);
    
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0)
        return sock;
    if (connect(sock, (struct sockaddr *) &ad, sizeof(ad)) < 0)
        return -1;
    return sock;
}
*/

// Open a TCP/IP socket and connect it to port port at host host.
static int
client_socket (char *host, int port)
{
  struct sockaddr_in address;
  struct hostent *hostent;
  int S;

  address.sin_addr.s_addr = inet_addr (host);
  if (address.sin_addr.s_addr == INADDR_NONE)
    {
      hostent = gethostbyname (host);
      if (hostent == 0)
	return -1;
      //address.sin_addr.s_addr = hostent->h_addr;
      memcpy(&address.sin_addr, hostent->h_addr, hostent->h_length);
    }

  address.sin_family = PF_INET;
  address.sin_port = htons (port);
  
  S = socket (PF_INET, SOCK_STREAM, 0);
  if (S == -1)
    {
      perror ("socket");
      return -1;
    }
  
  if (connect (S, (struct sockaddr *)&address, sizeof (address)) == -1)
    {
      perror ("connect");
      return -1;
    }

  return S;
}

Server *
open_server (char *host, int port, char *file)
{
  char buffer[1024];
  Server *server;
  int f, n;

  server = malloc (sizeof (Server));
  if (server == 0)
    {
      fprintf (stderr, "out of memory\n");
      return 0;
    }

  server->socket = client_socket (host, port);
  if (server->socket == -1)
    {
      fprintf (stderr, "can't open socket to %s\n", host);
      return 0;
    }

  f = open (file, O_RDONLY);
  if (f == -1)
    {
      fprintf (stderr, "can't open file %s", file);
      close (f);
      return 0;
    }

  while ((n = read (f, buffer, 1024)) > 0)
    write (server->socket, buffer, n);

  return server;
}

void
server_draw_area (Server *server, int x, int y, int w, int h,
		  int W, int H, int RPP)
{
  char command[100];

  sprintf (command, "A %d %d %d %d %d %d %d\n", x, y, w, h, W, H, RPP);
  write (server->socket, command, strlen (command));

  server->x0 = server->x = x;
  server->y0 = server->y = y;
  server->w = w;
  server->h = h;
  server->remaining_n = 0;
}

static int
server_get_next_byte (Server *server, unsigned char **data, size_t *n)
{
  if (server->remaining_n > 0)
    {
      server->remaining_n--;
      return server->remaining_buf[server->remaining_n];
    }
  else if (n > 0)
    {
      (*n)--;
      return *((*data)++);
    }
  else
    return 0;
}

int
server_receive_data (Server *server, Image_out_device *screen,
		     Image_out_device *file, unsigned char *data, size_t n)
{
  int r, g, b;

  if (n < 0)
    return SERVER_ERROR;

  if (n == 0)
    return SERVER_DONE;

  while (n >= 3)
    {
      r = server_get_next_byte (server, &data, &n);
      g = server_get_next_byte (server, &data, &n);
      b = server_get_next_byte (server, &data, &n);
      if (screen)
	screen->pixel (screen, server->x, server->y,
		       r / 255.0, g / 255.0, b / 255.0);
      if (file)
	file->pixel (file, server->x, server->y,
		     r / 255.0, g / 255.0, b / 255.0);
      server->x++;
      if (server->x - server->x0 == server->w)
	{
	  server->x = server->x0;
	  server->y++;
	  if (server->y - server->y0 == server->h)
	    return SERVER_DONE;
	}
    }

  while (n > 0)
    {
      server->remaining_buf[n - 1] = *data++;
      server->remaining_n++;
      n--;
    }
  
  return SERVER_NOT_DONE;
}

void
close_server (Server *server)
{
  if (server->remaining_n > 0)
    fprintf (stderr, "warning, server had remaining data\n");

  close (server->socket);
  free (server);
}
