// pix.c
// (C) 1997, 1998 lars brinkhoff

#include <stdio.h>
#include <stdlib.h>

#include "pix.h"

#ifdef HAVE_AALIB
#include "aa.h"
#endif

void read_image (struct Options *, void *);
void parse_options (int, char *[], struct Options *);
void *init (struct Options *);

extern char *optarg;
extern int optind, opterr, optopt;
extern getopt ();

int
main (int argc, char *argv[])
{
  struct Options options;
  void *info;

  parse_options (argc, argv, &options);
  info = options.init (&options);
  read_image (&options, info);
  options.wait (info);

  return 0;
}

void
read_image (struct Options *options, void *info)
{
  int x, y, i, j;
  int r, g, b;

  for (y = 0; options->height > 0 ? y < options->height : 1;
       y += options->interleave)
    for (x = 0; x < options->width; x += options->interleave)
      for (j = 0; j < options->interleave; j++)
	for (i = 0; i < options->interleave; i++)
	  {
	    r = getchar();
	    g = getchar();
	    b = getchar();
	    if (feof (stdin))
	      goto end;
	    options->pixel (info, x + i, y + j,
			    r / 255.0, g / 255.0, b / 255.0);
	  }
end:
}

void
parse_options (int argc, char *argv[], struct Options *options)
{ 
  int c;

  /* Defaults. */
  options->init = init;
  options->width = -1;
  options->height = -1;
  options->scale = 1;
  options->interleave = 1;
  options->me = argv[0];
  
  while ((c = getopt (argc, argv, "xaAs:w:h:i:")) != EOF)
    switch (c)
      {
#ifdef HAVE_X
      case 'x': options->init = init_x; break;
#endif
#ifdef HAVE_AALIB
      case 'A': options->init = init_aa; break;
#endif
      case 'a': options->init = init_ascii; break;
      case 's': options->scale = atoi (optarg); break;
      case 'w': options->width = atoi (optarg); break;
      case 'h': options->height = atoi (optarg); break;
      case 'i': options->interleave = atoi (optarg); break;
      }

  if (options->width < 0)
    {
      fprintf (stderr,
	       "%s: raw input width must be specified.\n", argv[0]);
      exit (1);
    }

  if (options->init != init_ascii && options->height < 0)
    {
      fprintf (stderr,
	       "%s: input height must be specified when not using ASCII.\n",
	       argv[0]);
      exit (1);
    }
}

void *
init (struct Options *options)
{
  void *info;
  
#ifdef HAVE_X
  info = init_x (options);
  if (info != 0)
    return info;
#endif
#ifdef HAVE_AA
  info = init_aa (options);
  if (info != 0)
    return info;
#endif
  return init_ascii (options);
}
